function [ dataOut, g, data0, mask ] = ...
    nolcos(k, M, m, times, useMask, findTube, accuracy)
% nolcos: solves example 1 from NOLCOS submission.
%
% [ data, g, data0, mask ] = nolcos(k, M, m, times, useMask, findTube,accuracy)
%  
% Solves Example 1 from the IFAC NOLCOS submission by 
%   Kurzhanski, Mitchell and Varaiya.  Involves backwards reachability for
%   an ellipsoidal target set within a convex constraint set.
%
% This example has dynamics
%
%       d/dt x_1 = x_2
%	d/dt x_2 = u		|u| \leq k
%
% The goal is to compute the backwards reachable set (u maximizing its size).
%
% Target set phi_f(x)^2 = (x - m)^T * M * (x - m) <= 1,
%   where M = M^T > 0 and m = [ 0, m_1 ]^T.
%
% Constraints phi_1(x) = |x_2| \leq 1.
%
% We are effectively calculating a value function which 
%   minimizes over u the maximum of
%     1) terminal payoff of phi_f
%     2) maximum phi_1 encountered along the trajectory
%
% Note that the implicit surfaces are not the usual signed distance functions
%   and zero sublevel sets.
%
% Parameters (all inputs have defaults):
%
%   k            Bound on control input magnitude (default = 1).
%   M		 Shape matrix for target ellipsoid (default = [ 8 0; 0 4 ]).
%   m            Center of target ellipsoid (default = [ 0; 0.7 ]).
%   times        Strictly increasing vector of times at which to return 
%                  data array (default = 1.0)
%   useMask      Boolean whether to apply constraints (default = 1).
%   findTube     Boolean.
%                  1             Calculate reach set over time (default).
%                  0             Calculate reach set at a particular time.
%   accuracy     Controls the order of approximations.
%                  'low'         Use odeCFL1 and upwindFirstFirst.
%                  'medium'      Use odeCFL2 and upwindFirstENO2 (default).
%                  'high'        Use odeCFL3 and upwindFirstENO3.
%                  'veryHigh'    Use odeCFL3 and upwindFirstWENO5.
%
%   data         Implicit surface function at t = times.
%                  If input times is a scalar, data is an array.
%                  If input times is a vector, data is a cell vector with
%                  one entry (an array) for each element of times.
%   g            Grid structure on which data was computed.
%   data0        Implicit surface function at t_0.
%   mask         Constraint implicit surface (negative inside constraints).
%
% Ian Mitchell, 4/14/04

%---------------------------------------------------------------------------
% Make sure we can see the kernel m-files.
run('../../addPathToKernel');

%---------------------------------------------------------------------------
% Input parameters
if(nargin < 1)
  k = 1;
end
if(nargin < 2)
  M = [ 8 0; 0 4 ];
end
if(nargin < 3)
  m = [ 0; 0.7 ];
end
if(nargin < 4)
  times = 1.0;
end
if(nargin < 5)
  useMask = 1;
end
if(nargin < 6)
  findTube = 1;
end
if(nargin < 7)
  accuracy = 'medium';
end

%---------------------------------------------------------------------------
% Integration parameters.
t0 = 0;                      % Start time.
singleStep = 0;              % Plot at each timestep (overrides tPlot).

% How many intermediate plots to produce?
plotSteps = 9;

% Period at which intermediate plots should be produced.
tPlot = (times(end) - t0) / (plotSteps - 1);

% How close (relative) do we need to get to tMax to be considered finished?
small = 100 * eps;

% Plot in separate subplots (set deleteLastPlot = 0 in this case)?
useSubplots = 0;

% Even if we are not using the mask, should we mask the initial conditions?
alwaysMaskIC = 0;

%---------------------------------------------------------------------------
% What level set should we view?
level = 1;

% Pause after each plot?
pauseAfterPlot = 0;

% Delete previous plot before showing next?
deleteLastPlot = 0;

%---------------------------------------------------------------------------
% Use periodic boundary conditions?
periodic = 0;

% Create the grid.
g.dim = 2;
g.min = [ -1.75; -1.00 ]; 
g.dx = 1/100;
if(periodic)
  g.max = [ +0.75; +1.50 ] - g.dx;
  g.bdry = @addGhostPeriodic;
else
  g.max = [ +0.75; +1.50 ];
  g.bdry = @addGhostExtrapolate;
end
g = processGrid(g);

%---------------------------------------------------------------------------
% What kind of display?
  switch(g.dim)
   case 1
    displayType = 'plot';
   case 2
    displayType = 'contour';    
   case 3
    displayType = 'surface';
   otherwise
    error('Default display type undefined for dimension %d', g.dim);
  end

%---------------------------------------------------------------------------
% Create target set phi_f (ellipsoid).  Note we are using the 1 level set.
data = zeros(size(g.xs{1}));
for i = 1 : g.dim
  for j = 1 : g.dim
    % not the most efficient, but the easiest to write
    data = data + (g.xs{i} - m(i)) .* M(i,j) .* (g.xs{j} - m(j));
  end
end
data = sqrt(data);
data0 = data;

%---------------------------------------------------------------------------
% Set up spatial approximation scheme.
schemeFunc = @termLaxFriedrichs;
schemeData.grid = g;
schemeData.hamFunc = @hamFunc;
schemeData.partialFunc = @partialFunc;
schemeData.dissFunc = @artificialDissipationGLF;
schemeData.k = k;

%---------------------------------------------------------------------------
% Set up time approximation scheme.
integratorOptions = odeCFLset('factorCFL', 0.5, 'stats', 'on');

if(singleStep)
  integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
end

%---------------------------------------------------------------------------
% Choose approximations at appropriate level of accuracy.
switch(accuracy)
 case 'low'
  schemeData.derivFunc = @upwindFirstFirst;
  integratorFunc = @odeCFL1;
 case 'medium'
  schemeData.derivFunc = @upwindFirstENO2;
  integratorFunc = @odeCFL2;
 case 'high'
  schemeData.derivFunc = @upwindFirstENO3;
  integratorFunc = @odeCFL3;
 case 'veryHigh'
  schemeData.derivFunc = @upwindFirstWENO5;
  integratorFunc = @odeCFL3;
 otherwise
  error('Unknown accuracy level %s', accuracy);
end

%---------------------------------------------------------------------------
% To compute the tube, we need to constrain the sign of the Hamiltonian.
if(findTube)
  innerData = schemeData;
  innerFunc = schemeFunc;

  clear schemeFunc schemeData;
  schemeFunc = @termRestrictUpdate;
  schemeData.innerFunc = innerFunc;
  schemeData.innerData = innerData;
  schemeData.positive = 0;
end

%---------------------------------------------------------------------------
% To apply a constraint, we need to modify the level set function
%   during integration by masking it.
if(useMask)

  % Mask should be negative (relative to level) inside the constraints.
  mask = abs(g.xs{2});

  % Ensure that the initial data satisfies the mask.
  data = max(data, mask);
  data0 = data;

  % Set up data required for the mask operation.
  %   Mask will be compared to vector form of data array used by integrator.
  schemeData.maskData = mask(:);
  schemeData.maskFunc = @max;

  % Let the integrator know what function to call.
  integratorOptions = odeCFLset(integratorOptions, ...
                                'postTimestep', @postTimestepMask);
else

  if(alwaysMaskIC)
    % Create the mask, but only use it to modify the initial conditions.
    mask = abs(g.xs{2});
    data = max(data, mask);
    data0 = data;
  else
    % Still need to provide output mask parameter for empty constraints.
    mask = 0;
  end

end

%---------------------------------------------------------------------------
% Initialize Display
f = figure;

% Set up subplot parameters if necessary.
if(useSubplots)
  rows = ceil(sqrt(plotSteps));
  cols = ceil(plotSteps / rows);
  plotNum = 1;
  subplot(rows, cols, plotNum);
end

h = visualizeLevelSet(g, data, displayType, level, [ 't = ' num2str(t0) ]);

hold on;
if(g.dim > 1)
  axis(g.axis);
  daspect([ 1 1 1 ]);
end

%---------------------------------------------------------------------------
% Loop until final time (subject to a little roundoff).
tNow = t0;
startTime = cputime;
if(length(times) > 1)
  dataOut = cell(length(times), 1);
end
for timeCount = 1 : length(times)
  while(times(timeCount) - tNow > small * times(end))

    % Reshape data array into column vector for ode solver call.
    y0 = data(:);

    % How far to step?
    tSpan = [ tNow, min(times(timeCount), tNow + tPlot) ];
  
    % Take a timestep.
    [ t y ] = feval(integratorFunc, schemeFunc, tSpan, y0,...
                    integratorOptions, schemeData);
    tNow = t(end);

    % Get back the correctly shaped data array
    data = reshape(y, g.shape);

    if(pauseAfterPlot)
      % Wait for last plot to be digested.
      pause;
    end

    % Get correct figure, and remember its current view.
    figure(f);
    figureView = view;
    
    % Delete last visualization if necessary.
    if(deleteLastPlot)
      delete(h);
    end
    
    % Move to next subplot if necessary.
    if(useSubplots)
      plotNum = plotNum + 1;
      subplot(rows, cols, plotNum);
    end
    
    % Create new visualization.
    h = visualizeLevelSet(g, data, displayType, level, [ 't = ' num2str(t0) ]);
    
    % Restore view.
    view(figureView);
  end

  if(length(times) == 1)
    dataOut = data;
  else
    dataOut{timeCount} = data;
  end

end

endTime = cputime;
fprintf('Total execution time %g seconds\n', endTime - startTime);



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function hamValue = hamFunc(t, data, deriv, schemeData)
% hamFunc: Hamiltonian for example 1 from the NOLCOS paper.
%
% hamValue = hamFunc(t, data, deriv, schemeData)
%
% Implements the hamFunc prototype for the NOLCOS example, which is a
%   simple double integrator with linear input.
%
% This example has dynamics (negated to reverse time)
%       d/dt x_1 = -x_2
%	d/dt x_2 = -u		|u| \leq k
%   where input authority is used to maximize the Hamiltonian
%	hamValue =  -x_2 * deriv{1} + k * |deriv{2}|
%
% Parameters:
%   t            Time at beginning of timestep (ignored).
%   data         Data array.
%   deriv	 Cell vector of the costate (\grad \phi).
%   schemeData	 A structure (see below).
%
%   hamValue	 The analytic hamiltonian.
%
% schemeData is a structure containing data specific to this Hamiltonian
%   For this function it contains the field(s):
%
%   .grid	 Grid structure.
%   .k           Scalar bound on control input magnitude.
%
% Ian Mitchell 4/14/04

checkStructureFields(schemeData, 'grid', 'k');

hamValue = -schemeData.grid.xs{2} .* deriv{1} + schemeData.k .* abs(deriv{2});



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function alpha = partialFunc(t, data, derivMin, derivMax, schemeData, dim)
% partialFunc: Hamiltonian partials function for NOLCOS paper example 1.
%
% alpha = partialFunc(t, data, derivMin, derivMax, schemeData, dim)
%
% Implements the partialFunc prototype for the NOLCOS example, which is a
%   simple double integrator with linear input.
%
% This example has dynamics (negated to reverse time)
%       d/dt x_1 = -x_2
%	d/dt x_2 = -u		|u| \leq k
%   where input authority is used to maximize the Hamiltonian.
%
% The magnitude of the extrema of the partials are
%       dim 1: abs(grid.xs{2})
%	dim 2: k
%
% Parameters:
%   t            Time at beginning of timestep (ignored).
%   data         Data array.
%   derivMin	 Cell vector of minimum values of the costate (\grad \phi).
%   derivMax	 Cell vector of maximum values of the costate (\grad \phi).
%   schemeData	 A structure (see below).
%   dim          Dimension in which the partial derivatives is taken.
%
%   alpha	 Maximum absolute value of the partial of the Hamiltonian
%		   with respect to the costate in dimension dim for the 
%                  specified range of costate values (O&F equation 5.12).
%		   Note that alpha can (and should) be evaluated separately
%		   at each node of the grid.
%
% schemeData is a structure containing data specific to this Hamiltonian
%   For this function it contains the field(s):
%
%   .grid	 Grid structure.
%   .k           Scalar bound on control input magnitude.
%
% Ian Mitchell 4/14/04

checkStructureFields(schemeData, 'grid', 'k');

switch(dim)
  case 1
    alpha = abs(schemeData.grid.xs{2});
  case 2
    alpha = abs(schemeData.k);
  otherwise
    error([ 'Unknown dimension: ' num2str(dim) ]);
end
